/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Netscape Portable Runtime (NSPR).
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998-2000
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "primpl.h"

#include <signal.h>

void _MD_EarlyInit(void)
{
    /*
     * Ignore FPE because coercion of a NaN to an int causes SIGFPE
     * to be raised.
     */
    struct sigaction act;

    act.sa_handler = SIG_IGN;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_RESTART;
    sigaction(SIGFPE, &act, 0);
}

PRWord *_MD_HomeGCRegisters(PRThread *t, int isCurrent, int *np)
{
#ifndef _PR_PTHREADS
    if (isCurrent) {
	(void) sigsetjmp(CONTEXT(t), 1);
    }
    *np = sizeof(CONTEXT(t)) / sizeof(PRWord);
    return (PRWord *) CONTEXT(t);
#else
	*np = 0;
	return NULL;
#endif
}

#if defined(_PR_PTHREADS)
/* JTL - 5/20/09 */
#if defined(ENABLE_SCTP)
PRInt32 _MD_Sctp_Recvmsg(
    int                  osfd,
    void                *buf,
    PRInt32              amount,
    PRNetAddr           *from,
    PRUint32            *fromlen,
    PRSctp_SndRcvInfo   *sinfo,
    PRInt32             *msg_flags)
{
    PRInt32 bytes;
    PRInt32 optval;
    PRUint32 length;
    PRSctp_ExtRcvInfo *serinfo=(PRSctp_ExtRcvInfo *)sinfo;
    struct sctp_extrcvinfo md_serinfo;
    struct sctp_sndrcvinfo *md_sinfo=(struct sctp_sndrcvinfo *)&md_serinfo;

    memset(&md_serinfo, 0, sizeof(md_serinfo));
    bytes = sctp_recvmsg(           osfd,
                                    buf,
                                    amount,
                 (struct sockaddr*) from,
                                    fromlen,
        (struct sctp_sndrcvinfo *) &md_serinfo,
                                    msg_flags);

    /* must check for SCTP_USE_EXT_RCVINFO */
    length = sizeof(optval);
    if(getsockopt(osfd, IPPROTO_SCTP, SCTP_USE_EXT_RCVINFO, (void *)&optval, &length)) {
        return -1;
    }

    if(optval) {          /* we are using extended receive info. */
        serinfo->serinfo_sinfo.sinfo_stream    = md_serinfo.sinfo_stream;
        serinfo->serinfo_sinfo.sinfo_ssn       = md_serinfo.sinfo_ssn;
        serinfo->serinfo_sinfo.sinfo_flags     = md_serinfo.sinfo_flags;
        serinfo->serinfo_sinfo.sinfo_pr_policy = md_serinfo.sinfo_pr_policy;
        serinfo->serinfo_sinfo.sinfo_ppid      = md_serinfo.sinfo_ppid;
        serinfo->serinfo_sinfo.sinfo_context   = md_serinfo.sinfo_context;
        serinfo->serinfo_sinfo.sinfo_pr_value  = md_serinfo.sinfo_timetolive;
        serinfo->serinfo_sinfo.sinfo_tsn       = md_serinfo.sinfo_tsn;
        serinfo->serinfo_sinfo.sinfo_cumtsn    = md_serinfo.sinfo_cumtsn;
        serinfo->serinfo_sinfo.sinfo_assoc_id  = md_serinfo.sinfo_assoc_id;
        serinfo->serinfo_next_flags            = md_serinfo.sreinfo_next_flags;
        serinfo->serinfo_next_stream           = md_serinfo.sreinfo_next_stream;
        serinfo->serinfo_next_aid              = md_serinfo.sreinfo_next_aid;
        serinfo->serinfo_next_length           = md_serinfo.sreinfo_next_length;
        serinfo->serinfo_next_ppid             = md_serinfo.sreinfo_next_ppid;
    }
    else {                /* we are not using extended receive info. */
        sinfo->sinfo_stream    = md_sinfo->sinfo_stream;
        sinfo->sinfo_ssn       = md_sinfo->sinfo_ssn;
        sinfo->sinfo_flags     = md_sinfo->sinfo_flags;
        sinfo->sinfo_pr_policy = md_sinfo->sinfo_pr_policy;
        sinfo->sinfo_ppid      = md_sinfo->sinfo_ppid;
        sinfo->sinfo_context   = md_sinfo->sinfo_context;
        sinfo->sinfo_pr_value  = md_sinfo->sinfo_timetolive;
        sinfo->sinfo_tsn       = md_sinfo->sinfo_tsn;
        sinfo->sinfo_cumtsn    = md_sinfo->sinfo_cumtsn;
        sinfo->sinfo_assoc_id  = md_sinfo->sinfo_assoc_id;
    }

    return bytes;
}

PRInt32 _MD_SCTPGetSocketOption(PRFileDesc *fd, PRSocketOptionData *data)
{
    PRStatus result;
    PRIntn rv;
    PRInt32 level, name;
    PRUint32 length;

    result = _PR_MapOptionName(data->option, &level, &name);
    if (PR_SUCCESS == result)
    {
        switch (data->option)
        {
            case PR_SockOpt_Sctp_Default_Send_Param:
            {
               /*
                * On FreeBSD we can't use PRSctp_SndRcvInfo because the native
                * structure contains padding for 64 bit alignment.
                */
                struct sctp_sndrcvinfo md_sndrcvinfo;
                length = sizeof(md_sndrcvinfo);
                memset(&md_sndrcvinfo, 0, length);
                rv = getsockopt(fd->secret->md.osfd, level, name, (void *)&md_sndrcvinfo,
                                &length);
                PR_ASSERT((-1 == rv) || (sizeof(md_sndrcvinfo) == length));
                data->value.sndrcvinfo.sinfo_stream    = md_sndrcvinfo.sinfo_stream;
                data->value.sndrcvinfo.sinfo_ssn       = md_sndrcvinfo.sinfo_ssn;
                data->value.sndrcvinfo.sinfo_flags     = md_sndrcvinfo.sinfo_flags;
                data->value.sndrcvinfo.sinfo_pr_policy = md_sndrcvinfo.sinfo_pr_policy;
                data->value.sndrcvinfo.sinfo_ppid      = md_sndrcvinfo.sinfo_ppid;
                data->value.sndrcvinfo.sinfo_context   = md_sndrcvinfo.sinfo_context;
                data->value.sndrcvinfo.sinfo_pr_value  = md_sndrcvinfo.sinfo_timetolive;
                data->value.sndrcvinfo.sinfo_tsn       = md_sndrcvinfo.sinfo_tsn;
                data->value.sndrcvinfo.sinfo_cumtsn    = md_sndrcvinfo.sinfo_cumtsn;
                data->value.sndrcvinfo.sinfo_assoc_id  = md_sndrcvinfo.sinfo_assoc_id;
                break;
            }
            case PR_SockOpt_Sctp_Events:
            {
                struct sctp_event_subscribe events;  /* Use native structure in case */
                length = sizeof(events);             /* this is FreeBSD 7.2.         */
                memset(&events, 0, length);
                rv = getsockopt(fd->secret->md.osfd, level, name, (void *)&events, &length);
                PR_ASSERT((-1 == rv) || (sizeof(events) == length));
                data->value.events.sctp_data_io_event =
                    events.sctp_data_io_event;
                data->value.events.sctp_association_event =
                    events.sctp_association_event;
                data->value.events.sctp_address_event =
                    events.sctp_address_event;
                data->value.events.sctp_send_failure_event =
                    events.sctp_send_failure_event;
                data->value.events.sctp_peer_error_event =
                    events.sctp_peer_error_event;
                data->value.events.sctp_shutdown_event =
                    events.sctp_shutdown_event;
                data->value.events.sctp_partial_delivery_event =
                    events.sctp_partial_delivery_event;
                data->value.events.sctp_adaptation_layer_event =
                    events.sctp_adaptation_layer_event;
                data->value.events.sctp_authentication_event =
                    events.sctp_authentication_event;
#if defined(SCTP_HAVE_DRY_EVENT_AND_RESET_EVENTS)
                data->value.events.sctp_sender_dry_event =
                    events.sctp_sender_dry_event;
#else
                data->value.events.sctp_sender_dry_event = 0;
#endif
                break;
            }
            case PR_SockOpt_Sctp_InitMsg:
            {
                struct sctp_initmsg md_initmsg;
                length = sizeof(md_initmsg);
                memset(&md_initmsg, 0, length);
                rv = getsockopt(fd->secret->md.osfd, level, name, (void *)&md_initmsg,
                                &length);
                PR_ASSERT((-1 == rv) || (sizeof(md_initmsg) == length));
                data->value.initmsg.sinit_num_ostreams   =
                    md_initmsg.sinit_num_ostreams;
                data->value.initmsg.sinit_max_instreams  =
                    md_initmsg.sinit_max_instreams;
                data->value.initmsg.sinit_max_attempts   =
                    md_initmsg.sinit_max_attempts;
                data->value.initmsg.sinit_max_init_timeo =
                    md_initmsg.sinit_max_init_timeo;
                break;
            }
            case PR_SockOpt_Sctp_AssocInfo:
            {
                struct sctp_assocparams md_assocparams;
                length = sizeof(md_assocparams);
                memset(&md_assocparams, 0, length);
                rv = getsockopt(fd->secret->md.osfd, level, name, (void *)&md_assocparams,
                                &length);
                PR_ASSERT((-1 == rv) || (sizeof(md_assocparams) == length));
                data->value.assocparams.sasoc_assoc_id =
                    md_assocparams.sasoc_assoc_id;
                data->value.assocparams.sasoc_asocmaxrxt =
                    md_assocparams.sasoc_asocmaxrxt;
                data->value.assocparams.sasoc_number_peer_destinations =
                    md_assocparams.sasoc_number_peer_destinations;
                data->value.assocparams.sasoc_peer_rwnd =
                    md_assocparams.sasoc_peer_rwnd;
                data->value.assocparams.sasoc_local_rwnd =
                    md_assocparams.sasoc_local_rwnd;
                data->value.assocparams.sasoc_cookie_life =
                    md_assocparams.sasoc_cookie_life;
                break;
            }
            case PR_SockOpt_Sctp_MaxSegment:
            {
                PRSctp_Assoc_Value sctp_maxseg;
                length = sizeof(sctp_maxseg);
                memset(&sctp_maxseg, 0, length);
                rv = getsockopt(
                    fd->secret->md.osfd, level, name, (void *)&sctp_maxseg, &length);
                PR_ASSERT((-1 == rv) || (sizeof(sctp_maxseg) == length));
                data->value.sctp_maxseg.assoc_id = sctp_maxseg.assoc_id;
                data->value.sctp_maxseg.assoc_value = sctp_maxseg.assoc_value;
                break;
            }
            case PR_SockOpt_Sctp_Status:
            {
                PRSctp_Status status;
                length = sizeof(status);
                memset(&status, 0, length);
                rv = getsockopt(
                    fd->secret->md.osfd, level, name, (void *)&status, &length);
                PR_ASSERT((-1 == rv) || (sizeof(status) == length));
                data->value.status.sstat_assoc_id  = status.sstat_assoc_id;
                data->value.status.sstat_state     = status.sstat_state;
                data->value.status.sstat_rwnd      = status.sstat_rwnd;
                data->value.status.sstat_unackdata = status.sstat_unackdata;
                data->value.status.sstat_penddata  = status.sstat_penddata;
                data->value.status.sstat_instrms   = status.sstat_instrms;
                data->value.status.sstat_outstrms  = status.sstat_outstrms;
                data->value.status.sstat_fragmentation_point =
                    status.sstat_fragmentation_point;
                data->value.status.sstat_primary   = status.sstat_primary;
                break;
            }
            default:
                PR_NOT_REACHED("Unknown socket option");
                break;
        }
        return rv;
    }
    return -1;    /* PR_MapOptionName returned PR_FAILURE */
}

PRInt32 _MD_SCTPSetSocketOption(PRFileDesc *fd, const PRSocketOptionData *data)
{
    PRStatus result;
    PRIntn rv;
    PRInt32 level, name;
    PRUint32 length;

    result = _PR_MapOptionName(data->option, &level, &name);
    if (PR_SUCCESS == result)
    {
        switch (data->option)
        {
            case PR_SockOpt_Sctp_Default_Send_Param:
            {
                PRSctp_SndRcvInfo sndrcvinfo;
                length = sizeof(sndrcvinfo);
                memset(&sndrcvinfo, 0, length);
                sndrcvinfo.sinfo_stream     = data->value.sndrcvinfo.sinfo_stream;
                sndrcvinfo.sinfo_ssn        = data->value.sndrcvinfo.sinfo_ssn;
                sndrcvinfo.sinfo_flags      = data->value.sndrcvinfo.sinfo_flags;
                sndrcvinfo.sinfo_pr_policy  = data->value.sndrcvinfo.sinfo_pr_policy;
                sndrcvinfo.sinfo_ppid       = data->value.sndrcvinfo.sinfo_ppid;
                sndrcvinfo.sinfo_context    = data->value.sndrcvinfo.sinfo_context;
                sndrcvinfo.sinfo_pr_value   = data->value.sndrcvinfo.sinfo_pr_value;
                sndrcvinfo.sinfo_tsn        = data->value.sndrcvinfo.sinfo_tsn;
                sndrcvinfo.sinfo_cumtsn     = data->value.sndrcvinfo.sinfo_cumtsn;
                sndrcvinfo.sinfo_assoc_id   = data->value.sndrcvinfo.sinfo_assoc_id;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&sndrcvinfo,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_Events:
            {
                struct sctp_event_subscribe events;  /* Use native structure in case */
                length = sizeof(events);             /* this is FreeBSD 7.2.         */
                memset(&events, 0, length);
                events.sctp_data_io_event =
                    data->value.events.sctp_data_io_event;
                events.sctp_association_event =
                    data->value.events.sctp_association_event;
                events.sctp_address_event =
                    data->value.events.sctp_address_event;
                events.sctp_send_failure_event =
                    data->value.events.sctp_send_failure_event;
                events.sctp_peer_error_event =
                    data->value.events.sctp_peer_error_event;
                events.sctp_shutdown_event =
                    data->value.events.sctp_shutdown_event;
                events.sctp_partial_delivery_event =
                    data->value.events.sctp_partial_delivery_event;
                events.sctp_adaptation_layer_event =
                    data->value.events.sctp_adaptation_layer_event;
                events.sctp_authentication_event =
                    data->value.events.sctp_authentication_event;
#if defined(SCTP_HAVE_DRY_EVENT_AND_RESET_EVENTS)
                events.sctp_sender_dry_event =
                    data->value.events.sctp_sender_dry_event;
#endif
                events.sctp_stream_reset_events = 0;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&events,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_InitMsg:
            {
                struct sctp_initmsg md_initmsg;
                length = sizeof(md_initmsg);
                memset(&md_initmsg, 0, length);
                md_initmsg.sinit_num_ostreams   =
                    data->value.initmsg.sinit_num_ostreams;
                md_initmsg.sinit_max_instreams  =
                    data->value.initmsg.sinit_max_instreams;
                md_initmsg.sinit_max_attempts   =
                    data->value.initmsg.sinit_max_attempts;
                md_initmsg.sinit_max_init_timeo =
                    data->value.initmsg.sinit_max_init_timeo;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&md_initmsg,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_AssocInfo:
            {
                struct sctp_assocparams md_assocparams;
                length = sizeof(md_assocparams);
                memset(&md_assocparams, 0, length);
                md_assocparams.sasoc_assoc_id =
                    data->value.assocparams.sasoc_assoc_id;
                md_assocparams.sasoc_asocmaxrxt =
                    data->value.assocparams.sasoc_asocmaxrxt;
                md_assocparams.sasoc_number_peer_destinations =
                    data->value.assocparams.sasoc_number_peer_destinations;
                md_assocparams.sasoc_peer_rwnd =
                    data->value.assocparams.sasoc_peer_rwnd;
                md_assocparams.sasoc_local_rwnd =
                    data->value.assocparams.sasoc_local_rwnd;
                md_assocparams.sasoc_cookie_life =
                    data->value.assocparams.sasoc_cookie_life;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&md_assocparams,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_MaxSegment:
            {
                PRSctp_Assoc_Value sctp_maxseg;
                length = sizeof(sctp_maxseg);
                memset(&sctp_maxseg, 0, length);
                sctp_maxseg.assoc_id = data->value.sctp_maxseg.assoc_id;
                sctp_maxseg.assoc_value = data->value.sctp_maxseg.assoc_value;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&sctp_maxseg,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_Set_Peer_Primary_Addr:
            {
                struct sctp_setpeerprim md_setpeerprim;
                length = sizeof(md_setpeerprim);
                memset(&md_setpeerprim, 0, length);
                md_setpeerprim.sspp_assoc_id = data->value.setpeerprim.sspp_assoc_id;
                md_setpeerprim.sspp_addr = data->value.setpeerprim.sspp_addr;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&md_setpeerprim,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_Set_Primary_Addr:
            {
                struct sctp_setprim md_setprim;
                length = sizeof(md_setprim);
                memset(&md_setprim, 0, length);
                md_setprim.ssp_assoc_id = data->value.setprim.ssp_assoc_id;
                md_setprim.ssp_addr = data->value.setprim.ssp_addr;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&md_setprim,
                                length);
                break;
            }
            default:
                PR_NOT_REACHED("Unknown socket option");
                break;
        }
        return rv;
    }
    return -1;    /* PR_MapOptionName returned PR_FAILURE - should never happen here. */
}
#endif /* ENABLE_SCTP */
#else /* ! _PR_PTHREADS */
void
_MD_SET_PRIORITY(_MDThread *thread, PRUintn newPri)
{
    return;
}

PRStatus
_MD_InitializeThread(PRThread *thread)
{
	return PR_SUCCESS;
}

PRStatus
_MD_WAIT(PRThread *thread, PRIntervalTime ticks)
{
    PR_ASSERT(!(thread->flags & _PR_GLOBAL_SCOPE));
    _PR_MD_SWITCH_CONTEXT(thread);
    return PR_SUCCESS;
}

PRStatus
_MD_WAKEUP_WAITER(PRThread *thread)
{
    if (thread) {
	PR_ASSERT(!(thread->flags & _PR_GLOBAL_SCOPE));
    }
    return PR_SUCCESS;
}

/* These functions should not be called for FreeBSD */
void
_MD_YIELD(void)
{
    PR_NOT_REACHED("_MD_YIELD should not be called for FreeBSD.");
}

PRStatus
_MD_CREATE_THREAD(
    PRThread *thread,
    void (*start) (void *),
    PRThreadPriority priority,
    PRThreadScope scope,
    PRThreadState state,
    PRUint32 stackSize)
{
    PR_NOT_REACHED("_MD_CREATE_THREAD should not be called for FreeBSD.");
	return PR_FAILURE;
}
#endif /* ! _PR_PTHREADS */
