/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Netscape Portable Runtime (NSPR).
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998-2000
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#include "primpl.h"


extern PRBool suspendAllOn;
extern PRThread *suspendAllThread;

extern void _MD_SET_PRIORITY(_MDThread *md, PRThreadPriority newPri);

PRIntervalTime _MD_Solaris_TicksPerSecond(void)
{
    /*
     * Ticks have a 10-microsecond resolution.  So there are
     * 100000 ticks per second.
     */
    return 100000UL;
}

/* Interval timers, implemented using gethrtime() */

PRIntervalTime _MD_Solaris_GetInterval(void)
{
    union {
	hrtime_t hrt;  /* hrtime_t is a 64-bit (long long) integer */
	PRInt64 pr64;
    } time;
    PRInt64 resolution;
    PRIntervalTime ticks;

    time.hrt = gethrtime();  /* in nanoseconds */
    /*
     * Convert from nanoseconds to ticks.  A tick's resolution is
     * 10 microseconds, or 10000 nanoseconds.
     */
    LL_I2L(resolution, 10000);
    LL_DIV(time.pr64, time.pr64, resolution);
    LL_L2UI(ticks, time.pr64);
    return ticks;
}

#ifdef _PR_PTHREADS
void _MD_EarlyInit(void)
{
}

PRWord *_MD_HomeGCRegisters(PRThread *t, PRIntn isCurrent, PRIntn *np)
{
	*np = 0;
	return NULL;
}

/* JTL - 5/20/09 */
#if defined(ENABLE_SCTP)
#include <string.h>
PRInt32 _MD_Sctp_Recvmsg(
    int                  osfd,
    void                *buf,
    PRInt32              amount,
    PRNetAddr           *from,
    PRUint32            *fromlen,
    PRSctp_SndRcvInfo   *sinfo,
    PRInt32             *msg_flags)
{
    PRInt32 bytes;
    struct sctp_sndrcvinfo md_sinfo;

    memset(&md_sinfo, 0, sizeof(md_sinfo));
    bytes = sctp_recvmsg(           osfd,
                                    buf,
                                    amount,
                 (struct sockaddr*) from,
                                    fromlen,
                                   &md_sinfo,
                                    msg_flags);

    sinfo->sinfo_stream    = md_sinfo.sinfo_stream;
    sinfo->sinfo_ssn       = md_sinfo.sinfo_ssn;
    sinfo->sinfo_flags     = md_sinfo.sinfo_flags;
    sinfo->sinfo_pr_policy = 0;          /* member doesn't exist on Solaris */
    sinfo->sinfo_ppid      = md_sinfo.sinfo_ppid;
    sinfo->sinfo_context   = md_sinfo.sinfo_context;
    sinfo->sinfo_pr_value  = md_sinfo.sinfo_timetolive;
    sinfo->sinfo_tsn       = md_sinfo.sinfo_tsn;
    sinfo->sinfo_cumtsn    = md_sinfo.sinfo_cumtsn;
    sinfo->sinfo_assoc_id  = md_sinfo.sinfo_assoc_id;

    return bytes;
}

PRInt32 _MD_SCTPGetSocketOption(PRFileDesc *fd, PRSocketOptionData *data)
{
    PRStatus result;
    PRIntn rv;
    PRInt32 level, name;
    PRUint32 length;

    result = _PR_MapOptionName(data->option, &level, &name);
    if (PR_SUCCESS == result)
    {
        switch (data->option)
        {
            case PR_SockOpt_Sctp_Default_Send_Param:
            {
                struct sctp_sndrcvinfo md_sndrcvinfo;
                length = sizeof(md_sndrcvinfo);
                memset(&md_sndrcvinfo, 0, length);
                rv = getsockopt(fd->secret->md.osfd, level, name, (void *)&md_sndrcvinfo,
                                &length);
                PR_ASSERT((-1 == rv) || (sizeof(md_sndrcvinfo) == length));
                data->value.sndrcvinfo.sinfo_stream    = md_sndrcvinfo.sinfo_stream;
                data->value.sndrcvinfo.sinfo_ssn       = md_sndrcvinfo.sinfo_ssn;
                data->value.sndrcvinfo.sinfo_flags     = md_sndrcvinfo.sinfo_flags;
                data->value.sndrcvinfo.sinfo_pr_policy = 0;
                data->value.sndrcvinfo.sinfo_ppid      = md_sndrcvinfo.sinfo_ppid;
                data->value.sndrcvinfo.sinfo_context   = md_sndrcvinfo.sinfo_context;
                data->value.sndrcvinfo.sinfo_pr_value  = md_sndrcvinfo.sinfo_timetolive;
                data->value.sndrcvinfo.sinfo_tsn       = md_sndrcvinfo.sinfo_tsn;
                data->value.sndrcvinfo.sinfo_cumtsn    = md_sndrcvinfo.sinfo_cumtsn;
                data->value.sndrcvinfo.sinfo_assoc_id  = md_sndrcvinfo.sinfo_assoc_id;
                break;
            }
            case PR_SockOpt_Sctp_Events:
            {
                struct sctp_event_subscribe md_events;
                length = sizeof(md_events);
                memset(&md_events, 0, length);
                rv = getsockopt(fd->secret->md.osfd, level, name, (void *)&md_events, &length);
                PR_ASSERT((-1 == rv) || (sizeof(md_events) == length));
                data->value.events.sctp_data_io_event =
                    md_events.sctp_data_io_event;
                data->value.events.sctp_association_event =
                    md_events.sctp_association_event;
                data->value.events.sctp_address_event =
                    md_events.sctp_address_event;
                data->value.events.sctp_send_failure_event =
                    md_events.sctp_send_failure_event;
                data->value.events.sctp_peer_error_event =
                    md_events.sctp_peer_error_event;
                data->value.events.sctp_shutdown_event =
                    md_events.sctp_shutdown_event;
                data->value.events.sctp_partial_delivery_event =
                    md_events.sctp_partial_delivery_event;
                data->value.events.sctp_adaptation_layer_event =
#if defined(SCTP_HAVE_ADAPTION)
                    md_events.sctp_adaption_layer_event;
#else
                    md_events.sctp_adaptation_layer_event;
#endif
                data->value.events.sctp_authentication_event = 0;
                data->value.events.sctp_sender_dry_event = 0;
                break;
            }
            case PR_SockOpt_Sctp_InitMsg:
            {
                PRSctp_InitMsg initmsg;
                length = sizeof(initmsg);
                memset(&initmsg, 0, length);
                rv = getsockopt(fd->secret->md.osfd, level, name, (void *)&initmsg,
                                &length);
                PR_ASSERT((-1 == rv) || (sizeof(initmsg) == length));
                data->value.initmsg.sinit_num_ostreams   =
                    initmsg.sinit_num_ostreams;
                data->value.initmsg.sinit_max_instreams  =
                    initmsg.sinit_max_instreams;
                data->value.initmsg.sinit_max_attempts   =
                    initmsg.sinit_max_attempts;
                data->value.initmsg.sinit_max_init_timeo =
                    initmsg.sinit_max_init_timeo;
                break;
            }
            case PR_SockOpt_Sctp_AssocInfo:
            {
                PRSctp_AssocParams assocparams;
                length = sizeof(assocparams);
                memset(&assocparams, 0, length);
                rv = getsockopt(fd->secret->md.osfd, level, name, (void *)&assocparams,
                                &length);
                PR_ASSERT((-1 == rv) || (sizeof(assocparams) == length));
                data->value.assocparams.sasoc_assoc_id =
                    assocparams.sasoc_assoc_id;
                data->value.assocparams.sasoc_asocmaxrxt =
                    assocparams.sasoc_asocmaxrxt;
                data->value.assocparams.sasoc_number_peer_destinations =
                    assocparams.sasoc_number_peer_destinations;
                data->value.assocparams.sasoc_peer_rwnd =
                    assocparams.sasoc_peer_rwnd;
                data->value.assocparams.sasoc_local_rwnd =
                    assocparams.sasoc_local_rwnd;
                data->value.assocparams.sasoc_cookie_life =
                    assocparams.sasoc_cookie_life;
                break;
            }
            case PR_SockOpt_Sctp_MaxSegment:
            {
               /*
                * Solaris doesn't support the sctp_assoc_value structure, but we can
                * ignore the assoc_id anyway since we only support 1-to-1 connections
                * and associations id's are only defined for 1-to-many connections.
                */
                PRUint32 sctp_maxseg;
                sctp_maxseg = 0;
                length = sizeof(sctp_maxseg);
                rv = getsockopt(
                    fd->secret->md.osfd, level, name, (void *)&sctp_maxseg, &length);
                PR_ASSERT((-1 == rv) || (sizeof(sctp_maxseg) == length));
                data->value.sctp_maxseg.assoc_value = sctp_maxseg;
                break;
            }
            case PR_SockOpt_Sctp_Status:
            {
                PRSctp_Status status;
                length = sizeof(status);
                memset(&status, 0, length);
                rv = getsockopt(
                    fd->secret->md.osfd, level, name, (void *)&status, &length);
                PR_ASSERT((-1 == rv) || (sizeof(status) == length));
                data->value.status.sstat_assoc_id  = status.sstat_assoc_id;
                data->value.status.sstat_state     = status.sstat_state;
                data->value.status.sstat_rwnd      = status.sstat_rwnd;
                data->value.status.sstat_unackdata = status.sstat_unackdata;
                data->value.status.sstat_penddata  = status.sstat_penddata;
                data->value.status.sstat_instrms   = status.sstat_instrms;
                data->value.status.sstat_outstrms  = status.sstat_outstrms;
                data->value.status.sstat_fragmentation_point =
                    status.sstat_fragmentation_point;
                data->value.status.sstat_primary   = status.sstat_primary;
                break;
            }
            default:
                PR_NOT_REACHED("Unknown socket option");
                break;
        }
        return rv;
    }
    return -1;    /* PR_MapOptionName returned PR_FAILURE */
}

PRInt32 _MD_SCTPSetSocketOption(PRFileDesc *fd, const PRSocketOptionData *data)
{
    PRStatus result;
    PRIntn rv;
    PRInt32 level, name;
    PRUint32 length;

    result = _PR_MapOptionName(data->option, &level, &name);
    if (PR_SUCCESS == result)
    {
        switch (data->option)
        {
            case PR_SockOpt_Sctp_Default_Send_Param:
            {
                struct sctp_sndrcvinfo md_sndrcvinfo;
                length = sizeof(md_sndrcvinfo);
                memset(&md_sndrcvinfo, 0, length);
                md_sndrcvinfo.sinfo_stream     = data->value.sndrcvinfo.sinfo_stream;
                md_sndrcvinfo.sinfo_ssn        = data->value.sndrcvinfo.sinfo_ssn;
                md_sndrcvinfo.sinfo_flags      = data->value.sndrcvinfo.sinfo_flags;
                /* pr_policy is not supported on Solaris. */
                md_sndrcvinfo.sinfo_ppid       = data->value.sndrcvinfo.sinfo_ppid;
                md_sndrcvinfo.sinfo_context    = data->value.sndrcvinfo.sinfo_context;
                md_sndrcvinfo.sinfo_timetolive = data->value.sndrcvinfo.sinfo_pr_value;
                md_sndrcvinfo.sinfo_tsn        = data->value.sndrcvinfo.sinfo_tsn;
                md_sndrcvinfo.sinfo_cumtsn     = data->value.sndrcvinfo.sinfo_cumtsn;
                md_sndrcvinfo.sinfo_assoc_id   = data->value.sndrcvinfo.sinfo_assoc_id;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&md_sndrcvinfo,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_Events:
            {
                struct sctp_event_subscribe md_events;
                length = sizeof(md_events);
                memset(&md_events, 0, length);
                md_events.sctp_data_io_event =
                    data->value.events.sctp_data_io_event;
                md_events.sctp_association_event =
                    data->value.events.sctp_association_event;
                md_events.sctp_address_event =
                    data->value.events.sctp_address_event;
                md_events.sctp_send_failure_event =
                    data->value.events.sctp_send_failure_event;
                md_events.sctp_peer_error_event =
                    data->value.events.sctp_peer_error_event;
                md_events.sctp_shutdown_event =
                    data->value.events.sctp_shutdown_event;
                md_events.sctp_partial_delivery_event =
                    data->value.events.sctp_partial_delivery_event;
#if defined(SCTP_HAVE_ADAPTION)
                md_events.sctp_adaption_layer_event =
#else
                md_events.sctp_adaptation_layer_event =
#endif
                    data->value.events.sctp_adaptation_layer_event;
                /* authentication_layer_event not support on Solaris */
                /* sender_dry_event not support on Solaris */
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&md_events,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_InitMsg:
            {
                PRSctp_InitMsg initmsg;
                length = sizeof(initmsg);
                memset(&initmsg, 0, length);
                initmsg.sinit_num_ostreams   =
                    data->value.initmsg.sinit_num_ostreams;
                initmsg.sinit_max_instreams  =
                    data->value.initmsg.sinit_max_instreams;
                initmsg.sinit_max_attempts   =
                    data->value.initmsg.sinit_max_attempts;
                initmsg.sinit_max_init_timeo =
                    data->value.initmsg.sinit_max_init_timeo;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&initmsg,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_AssocInfo:
            {
                PRSctp_AssocParams assocparams;
                length = sizeof(assocparams);
                memset(&assocparams, 0, length);
                assocparams.sasoc_assoc_id =
                    data->value.assocparams.sasoc_assoc_id;
                assocparams.sasoc_asocmaxrxt =
                    data->value.assocparams.sasoc_asocmaxrxt;
                assocparams.sasoc_number_peer_destinations =
                    data->value.assocparams.sasoc_number_peer_destinations;
                assocparams.sasoc_peer_rwnd =
                    data->value.assocparams.sasoc_peer_rwnd;
                assocparams.sasoc_local_rwnd =
                    data->value.assocparams.sasoc_local_rwnd;
                assocparams.sasoc_cookie_life =
                    data->value.assocparams.sasoc_cookie_life;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&assocparams,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_MaxSegment:
            {
                PRUint32 sctp_maxseg;
                sctp_maxseg = data->value.sctp_maxseg.assoc_value;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&sctp_maxseg,
                                sizeof(sctp_maxseg));
                break;
            }
            case PR_SockOpt_Sctp_Set_Peer_Primary_Addr:
            {
                PRSctp_SetPeerPrim setpeerprim;
                length = sizeof(setpeerprim);
                memset(&setpeerprim, 0, length);
                setpeerprim.sspp_assoc_id = data->value.setpeerprim.sspp_assoc_id;
                setpeerprim.sspp_addr = data->value.setpeerprim.sspp_addr;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&setpeerprim,
                                length);
                break;
            }
            case PR_SockOpt_Sctp_Set_Primary_Addr:
            {
                PRSctp_SetPrim setprim;
                length = sizeof(setprim);
                memset(&setprim, 0, length);
                setprim.ssp_assoc_id = data->value.setprim.ssp_assoc_id;
                setprim.ssp_addr = data->value.setprim.ssp_addr;
                rv = setsockopt(fd->secret->md.osfd, level, name, (void *)&setprim,
                                length);
                break;
            }
            default:
                PR_NOT_REACHED("Unknown socket option");
                break;
        }
        return rv;
    }
    return -1;    /* PR_MapOptionName returned PR_FAILURE - should never happen here. */
}
#endif /* ENABLE_SCTP */
#endif /* _PR_PTHREADS */

#if defined(_PR_LOCAL_THREADS_ONLY)

void _MD_EarlyInit(void)
{
}

void _MD_SolarisInit()
{
    _PR_UnixInit();
}

void
_MD_SET_PRIORITY(_MDThread *thread, PRThreadPriority newPri)
{
    return;
}

PRStatus
_MD_InitializeThread(PRThread *thread)
{
	return PR_SUCCESS;
}

PRStatus
_MD_WAIT(PRThread *thread, PRIntervalTime ticks)
{
    PR_ASSERT(!(thread->flags & _PR_GLOBAL_SCOPE));
    _PR_MD_SWITCH_CONTEXT(thread);
    return PR_SUCCESS;
}

PRStatus
_MD_WAKEUP_WAITER(PRThread *thread)
{
	PR_ASSERT((thread == NULL) || (!(thread->flags & _PR_GLOBAL_SCOPE)));
    return PR_SUCCESS;
}

/* These functions should not be called for Solaris */
void
_MD_YIELD(void)
{
    PR_NOT_REACHED("_MD_YIELD should not be called for Solaris");
}

PRStatus
_MD_CREATE_THREAD(
    PRThread *thread,
    void (*start) (void *),
    PRThreadPriority priority,
    PRThreadScope scope,
    PRThreadState state,
    PRUint32 stackSize)
{
    PR_NOT_REACHED("_MD_CREATE_THREAD should not be called for Solaris");
	return(PR_FAILURE);
}

#ifdef USE_SETJMP
PRWord *_MD_HomeGCRegisters(PRThread *t, int isCurrent, int *np)
{
    if (isCurrent) {
		(void) setjmp(CONTEXT(t));
    }
    *np = sizeof(CONTEXT(t)) / sizeof(PRWord);
    return (PRWord *) CONTEXT(t);
}
#else
PRWord *_MD_HomeGCRegisters(PRThread *t, PRIntn isCurrent, PRIntn *np)
{
    if (isCurrent) {
		(void) getcontext(CONTEXT(t));
    }
    *np = NGREG;
    return (PRWord*) &t->md.context.uc_mcontext.gregs[0];
}
#endif  /* USE_SETJMP */

#endif  /* _PR_LOCAL_THREADS_ONLY */

#ifndef _PR_PTHREADS
#if defined(i386) && defined(SOLARIS2_4)
/* 
 * Because clock_gettime() on Solaris/x86 2.4 always generates a
 * segmentation fault, we use an emulated version _pr_solx86_clock_gettime(),
 * which is implemented using gettimeofday().
 */

int
_pr_solx86_clock_gettime(clockid_t clock_id, struct timespec *tp)
{
    struct timeval tv;

    if (clock_id != CLOCK_REALTIME) {
	errno = EINVAL;
	return -1;
    }

    gettimeofday(&tv, NULL);
    tp->tv_sec = tv.tv_sec;
    tp->tv_nsec = tv.tv_usec * 1000;
    return 0;
}
#endif  /* i386 && SOLARIS2_4 */
#endif  /* ! _PR_PTHREADS */
